
-- ************
-- * UI. Hint *
-- ************

--[[
- Import libraries, other files.
- Create hint library.

DToX, 2013-2015
Licensed under the same terms as Lua itself.
--]]





-- *****************
-- * Documentation *
-- *****************
--[[
To create and show hint use Hint( <object>, <text> ).
Example:
   function numberField:MouseEnter()
      ...
      Hint( self, "Input number." );
      ...
   end

Hint is removed automatically, if there is no MouseLeave event.
But if there is one, you have to add Hint.Remove() to your event's function
or set setting "AutoRemove" to true.
Example:
   1)
   function label:MouseLeave()
      ...
      Hint.Remove();
      ...
   end
   2) Or
   Hint.AutoRemove = true;

Warning. AutoRemove will result in a very small and very rare memory leak.
It shouldn't matter. Just for your information.

For possible visuals settings check under "Library's object and settings" comment.

--]]





-- import "Turbine";
-- import "Turbine.UI";
import "Turbine.UI.Lotro";



-- Dependencies
local Text; -- Library: "Text.lua"

-- Shorthand
local color = Turbine.UI.Color;
local align = Turbine.UI.ContentAlignment;
local font = Turbine.UI.Lotro.Font;
local wl = Turbine.Shell.WriteLine;

-- Defaults
local FONT = "Verdana14";
local FONT_COLOR = "DarkKhaki";
local TEXT_WIDTH = 200;
local OUTER_OFFSET_V = 8;
local INNER_OFFSET = 7;
-- Border images
local CORNER_WIDTH = 10;
local CORNER_HEIGHT = 10;
local BORDER_WIDTH = 7;
local BORDER_HEIGHT = 7;



-- Library's object and settings
Hint = {};
Hint.Font = FONT;
   -- Name or code. Example: "Verdana14" or 1107296257.
Hint.FontColor = FONT_COLOR;
   -- Name or code. Example: "DarkKhaki" or Turbine.UI.Color( 0.5, 0.5, 0.5 ).
Hint.TextWidth = TEXT_WIDTH;
   --[[
   Number or "Auto".
   Number = limit width, multiline.
   "Auto" = fit text. Maximum width is limited to the screen's width.
   --]]
Hint.InnerOffset = INNER_OFFSET;
   -- Hint's borders <-> hint's text.
Hint.OuterOffset = OUTER_OFFSET_V;
   -- Object's border <-> hint's top/bottom border.
Hint.AutoRemove = false;
   --[[
   If set to true, hint will automatically close itself. Adds to the MouseLeave event.
   Note. If there is no MouseLeave event, one will be created regardless of the setting.
   Warning. Will result in a rare and small memory leak.
   --]]

-- To use Hint instead of Hint.New
local mt = {};
mt.__call = function( _, ... )
   return Hint.New( ... );
end
setmetatable( Hint, mt );





-- Forward declarations
--
local carrier;
local hintWidth;
local hintHeight;



-- Positions hint
local function positionHint( carrier, obj )
   local screenWidth, screenHeight = Turbine.UI.Display:GetSize();
   local objWidth, objHeight = obj:GetSize();
   local screenX, screenY = obj:PointToScreen( 0, objHeight );
   local outerOffsetH = ( hintWidth - objWidth ) / 2;
   local outerOffsetV = Hint.OuterOffset;
   local left = screenX - outerOffsetH;

   -- Hint is larger than object. Correct horizontal position
   if ( outerOffsetH > 0 ) then
      -- Correct hint's right position
      if ( left + hintWidth > screenWidth ) then
         -- left = screenX + objWidth - hintWidth; -- Same line as object's right border
         left = screenWidth - hintWidth;
      end
      -- Correct hint's left position
      if ( left < 0 ) then
         -- left = screenX; -- Same line as object's left border
         left = 0;
      end
   end

   local top = screenY + outerOffsetV;
   -- Correct hint's vertical position
   if ( top + hintHeight > screenHeight ) then
      top = screenY - objHeight - outerOffsetV - hintHeight;
   end
   -- Check new top
   if ( top < 0 ) then
      top = 0;
   end

   carrier:SetPosition( left, top );
end

-- Adds one part of the border's images
local function addHintBorderImage( parent, imageID, width, height )
   local obj = Turbine.UI.Control();
   obj:SetParent( parent );
   obj:SetBackground( imageID );
   obj:SetBlendMode( 4 ); -- Blend with background. Also works: 5, 7
   obj:SetSize( width, height );
   return obj;
end

-- Adds border's images
local function addHintBorder( parent )
   --[[ Background and corners
   Width*Height

   Old (before U12). All - 19*19
      Top left corner - 0x41000133
      Top right corner - 0x41000139
      Bottom right corner - 0x41000137
      Bottom left corner - 0x41000135

      Top border - 0x4100013A; 2*3
      Bottom border - 0x41000136; 2*3
      Left border -  0x41000134; 3*2
      Right border - 0x41000138; 3*2

      Background - 0x4100013B; 128*128

   New (U12). All - 10*10
      Top left corner - 0x4114061F
      Top right corner - 0x41140618
      Bottom right corner - 0x41140620
      Bottom left corner - 0x41140619

      Top border - 0x4114061D; 10*7
      Bottom border - 0x4114061C; 10*7
      Left border -  0x4114061A; 7*10
      Right border - 0x4114061E; 7*10

      Background - 0x4114061B; 283*382
   --]]

   -- Calculate size
   local parentW, parentH = parent:GetSize();
   local borderHeight = parentH - CORNER_HEIGHT * 2;
   local borderWidth = parentW - CORNER_WIDTH * 2;

   -- 1
   local topLeftCorner = addHintBorderImage( parent, 0x4114061F, CORNER_WIDTH, CORNER_HEIGHT );
   topLeftCorner:SetPosition( 0, 0 );
   -- 2
   local topBorder = addHintBorderImage( parent, 0x4114061D, borderWidth, BORDER_HEIGHT );
   topBorder:SetPosition( CORNER_WIDTH, 0 );
   -- 3
   local topRightCorner = addHintBorderImage( parent, 0x41140618, CORNER_WIDTH, CORNER_HEIGHT );
   topRightCorner:SetPosition( parentW - CORNER_WIDTH, 0 );
   -- 4
   local leftBorder = addHintBorderImage( parent, 0x4114061A, BORDER_WIDTH, borderHeight );
   leftBorder:SetPosition( 0, CORNER_HEIGHT );
   -- 5
   local rightBorder = addHintBorderImage( parent, 0x4114061E, BORDER_WIDTH, borderHeight );
   rightBorder:SetPosition( parentW - BORDER_WIDTH, CORNER_HEIGHT );
   -- 6
   local bottomLeftCorner = addHintBorderImage( parent, 0x41140619, CORNER_WIDTH, CORNER_HEIGHT );
   bottomLeftCorner:SetPosition( 0, parentH - CORNER_HEIGHT );
   -- 7
   local bottomBorder = addHintBorderImage( parent, 0x4114061C, borderWidth, BORDER_HEIGHT );
   bottomBorder:SetPosition( CORNER_WIDTH, parentH - BORDER_HEIGHT );
   -- 8
   local bottomRightCorner = addHintBorderImage( parent, 0x41140620, CORNER_WIDTH, CORNER_HEIGHT );
   bottomRightCorner:SetPosition( parentW - CORNER_WIDTH, parentH - CORNER_HEIGHT );
end

-- Creates and shows object's hint.
-- If object doesn't have a MouseLeave event, adds one, that removes hint
Hint.New = function( obj, text )

   -- Check dependencies
   if ( not Text ) then
      if ( _G.DToX_Lib.Text ) then
      Text = _G.DToX_Lib.Text;
      else
         wl( ">> Hint. Error. Please, import 'Text.lua' library." );
      end
   end

   -- Don't create if already have one
   if ( carrier ) then
      return;
   end

   -- No object. Show error
   if ( not obj ) then
      wl( ">> Hint. Error. Please, set the object." );
      Hint.Remove();
      return;
   end


   local innerOffset = Hint.InnerOffset;
   local textWidth = Hint.TextWidth;
   -- Why would you...
   if ( not text or text == "" ) then
      text = "Well, it works.\n\n<rgb=#FFFF00>Now, please, set your text.</rgb>";
   end


   -- Holds full construction
   carrier = Turbine.UI.Window();
   carrier:SetVisible( true );
   carrier:SetZOrder( 100 );
   carrier:SetOpacity( 0 );

   local carrierBackground = Turbine.UI.Control();
   carrierBackground:SetParent( carrier );
   carrierBackground:SetBackground( 0x4114061B );
   carrierBackground:SetPosition( 2, 2 );

   -- Set and position text
   local c = color[Hint.FontColor] or Hint.FontColor;
   local f = font[Hint.Font] or Hint.Font;
   local label = Turbine.UI.Label();
   label:SetParent( carrier );
   label:SetZOrder( 1 ); -- To be above background images
   label:SetPosition( innerOffset, innerOffset );
   label:SetForeColor( c );
   label:SetTextAlignment( align.MiddleCenter );
   label:SetMarkupEnabled( true );
   label:SetFont( f );
   label:SetText( text );

   -- Calculate text's size
   local textW, textH = Text:getSize( text, f, textWidth );

   -- No or empty text. Error is shown in the hint
   if ( not textW ) then
      Hint.Remove();
      return;
   end

   -- Shrink text's width to the screen's width
   if ( Hint.TextWidth == "Auto" ) then
      local screenWidth = Turbine.UI.Display:GetWidth();
      local hintWidth = textW + innerOffset * 2;
      if ( hintWidth > screenWidth ) then
         textW, textH = Text:getSize( text, f, screenWidth );
      end
   end

   -- Set sizes
   hintWidth = textW + innerOffset * 2;
   hintHeight = textH + innerOffset * 2;
   carrier:SetSize( hintWidth, hintHeight );
   carrierBackground:SetSize( hintWidth - 2 * 2, hintHeight - 2 * 2 );
   label:SetSize( textW, textH );

   -- Add border, check position, show hint
   addHintBorder( carrier ); -- +9 KB. Problem: SetParent
   positionHint( carrier, obj );
   carrier:SetOpacity( 1 );

   -- Add hint's removal
   if ( not obj.MouseLeave ) then
      obj.MouseLeave = Hint.Remove;
   elseif ( Hint.AutoRemove ) then -- Rare and small memory leak...
      local orgMouseLeave = obj.MouseLeave;
      function obj.MouseLeave( self, ... )
         orgMouseLeave( self, ... );
         Hint.Remove();
      end
   end

end

-- Removes hint
Hint.Remove = function()
   if ( carrier ) then
      carrier:GetControls():Clear();
      carrier = nil;
   end
end





_G.DToX_Lib = _G.DToX_Lib or {};
_G.DToX_Lib.Hint = Hint;
